from collections import defaultdict
from time import time
from typing import Optional, Union
from logging import getLogger

import bw2calc as bc
import bw2data as bd
import numpy as np
import pandas as pd
from stats_arrays import MCRandomNumberGenerator

from activity_browser.mod import bw2data as bd

from .manager import MonteCarloParameterManager

log = getLogger(__name__)


class MonteCarloLCA(object):
    """A Monte Carlo LCA for multiple reference flows and methods loaded from a calculation setup."""

    def __init__(self, cs_name):
        if cs_name not in bd.calculation_setups:
            raise ValueError("{} is not a known `calculation_setup`.".format(cs_name))

        self.cs_name = cs_name
        self.cs = bd.calculation_setups[cs_name]
        self.seed = None
        self.cf_rngs = {}
        self.CF_rng_vectors = {}
        self.include_technosphere = True
        self.include_biosphere = True
        self.include_cfs = False
        # Needs substantial rework for BW 2.5
        self.include_parameters = False
        # self.param_rng = None
        self.param_cols = ["row", "col", "type"]

        # self.tech_rng: Optional[Union[MCRandomNumberGenerator, np.ndarray]] = None
        # self.bio_rng: Optional[Union[MCRandomNumberGenerator, np.ndarray]] = None
        # self.cf_rng: Optional[Union[MCRandomNumberGenerator, np.ndarray]] = None

        # reference flows
        self.func_units = self.cs["inv"]
        self.rev_fu_index = {i: fu for i, fu in enumerate(self.func_units)}
        self.func_units_multilca = {str(i): fu for i, fu in enumerate(self.func_units)}

        # activities
        self.activity_keys = [list(fu.keys())[0] for fu in self.func_units]
        self.activity_index = {
            key: index for index, key in enumerate(self.activity_keys)
        }
        self.rev_activity_index = {
            index: key for index, key in enumerate(self.activity_keys)
        }

        # methods
        self.methods = list(self.cs["ia"])
        self.method_index = {m: i for i, m in enumerate(self.methods)}
        self.rev_method_index = {i: m for i, m in enumerate(self.methods)}

        # GSA calculation variables
        self.A_matrices = list()
        self.B_matrices = list()
        self.CF_dict = defaultdict(list)
        self.parameter_exchanges = list()
        self.parameters = list()
        self.parameter_data = defaultdict(dict)

        self.results = list()

        # Not needed - always call .calculate?
        # self.lca = self.construct_lca(
        #     demands=self.func_units,
        #     method_config={"impact_categories": self.methods},
        #     technosphere=self.include_technosphere,
        #     biosphere=self.include_biosphere,
        #     characterization=self.include_cfs,
        # )

    def construct_lca(
        self,
        demands: dict,
        method_config: dict,
        technosphere: bool = True,
        biosphere: bool = True,
        characterization: bool = True,
        seed_override: Optional[int] = None,
    ) -> bc.MultiLCA:
        log.info(f"Monte Carlo demands: {demands}")
        log.info(f"Monte Carlo impact categories: {method_config}")
        demands = {
            index: {bd.get_activity(k).id: v for k, v in fu.items()}
            for index, fu in demands.items()
        }
        data_objs = bd.get_multilca_data_objs(
            functional_units=demands,
            method_config=method_config
        )
        return bc.MultiLCA(
            demands=demands,
            method_config=method_config,
            data_objs=data_objs,
            selective_use={
                "technosphere_matrix": {"use_distributions": technosphere},
                "biosphere_matrix": {"use_distributions": biosphere},
                "characterization_matrix": {"use_distributions": characterization},
            },
            seed_override=seed_override,
        )

    # def unify_param_exchanges(self, data: np.ndarray) -> np.ndarray:
    #     """Convert an array of parameterized exchanges from input/output keys
    #     into row/col values using dicts generated in bw.LCA object.

    #     If any given exchange does not exist in the current LCA matrix,
    #     it will be dropped from the returned array.
    #     """

    #     def key_to_rowcol(x) -> Optional[tuple]:
    #         if x["type"] in [0, 1]:
    #             row = self.lca.activity_dict.get(x["input"], None)
    #             col = self.lca.product_dict.get(x["output"], None)
    #         else:
    #             row = self.lca.biosphere_dict.get(x["input"], None)
    #             col = self.lca.activity_dict.get(x["output"], None)
    #         # if either the row or the column is None, return np.NaN.
    #         if row is None or col is None:
    #             return None
    #         return row, col, x["type"], x["amount"]

    #     # Convert the data and store in a new array, dropping Nones.
    #     converted = (key_to_rowcol(d) for d in data)
    #     unified = np.array(
    #         [x for x in converted if x is not None],
    #         dtype=[("row", "<u4"), ("col", "<u4"), ("type", "u1"), ("amount", "<f4")],
    #     )
    #     return unified

    # def load_data(self) -> None:
    #     """Constructs the random number generators for all of the matrices that
    #     can be altered by uncertainty.

    #     If any of these uncertain calculations are not included, the initial
    #     amounts of the 'params' matrices are used in place of generating
    #     a vector
    #     """
    #     self.lca.load_lci_data()

    #     self.tech_rng = (
    #         MCRandomNumberGenerator(self.lca.tech_params, seed=self.seed)
    #         if self.include_technosphere
    #         else self.lca.tech_params["amount"].copy()
    #     )
    #     self.bio_rng = (
    #         MCRandomNumberGenerator(self.lca.bio_params, seed=self.seed)
    #         if self.include_biosphere
    #         else self.lca.bio_params["amount"].copy()
    #     )

    #     if self.lca.lcia:
    #         self.cf_rngs = (
    #             {}
    #         )  # we need as many cf_rng as impact categories, because they are of different size
    #         for m in self.methods:
    #             self.lca.switch_method(m)
    #             self.lca.load_lcia_data()
    #             self.cf_rngs[m] = (
    #                 MCRandomNumberGenerator(self.lca.cf_params, seed=self.seed)
    #                 if self.include_cfs
    #                 else self.lca.cf_params["amount"].copy()
    #             )
    #     # Construct the MC parameter manager
    #     if self.include_parameters:
    #         self.param_rng = MonteCarloParameterManager(seed=self.seed)

    #     (
    #         self.lca.activity_dict_rev,
    #         self.lca.product_dict_rev,
    #         self.lca.biosphere_dict_rev,
    #     ) = self.lca.reverse_dict()

    def calculate(self, iterations: int = 10, seed: Optional[int] = None, **kwargs):
        """Main calculate method for the MC LCA class, allows fine-grained control
        over which uncertainties are included when running MC sampling.
        """
        start = time()
        self.iterations = iterations
        self.seed = seed or bc.utils.get_seed()
        self.include_technosphere = kwargs.get("technosphere", True)
        self.include_biosphere = kwargs.get("biosphere", True)
        self.include_cfs = kwargs.get("cf", True)
        self.include_parameters = kwargs.get("parameters", True)

        self.lca = self.construct_lca(
            demands=self.func_units_multilca,
            method_config={"impact_categories": self.methods},
            technosphere=self.include_technosphere,
            biosphere=self.include_biosphere,
            characterization=self.include_cfs,
        )
        self.lca.lci()
        self.lca.lcia()
        self.lca.keep_first_iteration_flag = True

        self.results = np.zeros((iterations, len(self.func_units), len(self.methods)))

        # Reset GSA variables to empty.
        self.A_matrices = list()
        self.B_matrices = list()
        self.CF_dict = defaultdict(list)
        self.parameter_exchanges = list()
        self.parameters = list()

        # Prepare GSA parameter schema:
        if self.include_parameters:
            raise ValueError("Parameter sampling not currently supported")
            # self.parameter_data = self.param_rng.extract_active_parameters(self.lca)
            # # Add a values field to handle all the sampled parameter values.
            # for k in self.parameter_data:
            #     self.parameter_data[k]["values"] = []

        for iteration in range(iterations):
            next(self.lca)
            # tech_vector = (
            #     self.tech_rng.next() if self.include_technosphere else self.tech_rng
            # )
            # bio_vector = self.bio_rng.next() if self.include_biosphere else self.bio_rng
            # if self.include_parameters:
            #     # Convert the input/output keys into row/col keys, and then match them against
            #     # the tech_ and bio_params
            #     data = self.param_rng.next()
            #     param_exchanges = self.unify_param_exchanges(data)

            #     # Select technosphere subset from param_exchanges.
            #     subset = param_exchanges[np.isin(param_exchanges["type"], [0, 1])]
            #     # Create index of where to insert new values from tech_params array.
            #     idx = np.argwhere(
            #         np.isin(
            #             self.lca.tech_params[self.param_cols], subset[self.param_cols]
            #         )
            #     ).flatten()
            #     # Construct unique array of row+col+type combinations
            #     uniq = np.unique(self.lca.tech_params[idx][self.param_cols])
            #     # Use the unique array to sort the subset (ensures values
            #     # are inserted at the correct index)
            #     sort_idx = np.searchsorted(uniq, subset[self.param_cols])
            #     # Finally, insert the sorted subset amounts into the tech_vector
            #     # at the correct indexes.
            #     tech_vector[idx] = subset[sort_idx]["amount"]
            #     # Repeat the above, but for the biosphere array.
            #     subset = param_exchanges[param_exchanges["type"] == 2]
            #     idx = np.argwhere(
            #         np.isin(
            #             self.lca.bio_params[self.param_cols], subset[self.param_cols]
            #         )
            #     ).flatten()
            #     uniq = np.unique(self.lca.bio_params[idx][self.param_cols])
            #     sort_idx = np.searchsorted(uniq, subset[self.param_cols])
            #     bio_vector[idx] = subset[sort_idx]["amount"]

            #     # Store parameter data for GSA
            #     self.parameter_exchanges.append(param_exchanges)
            #     self.parameters.append(self.param_rng.parameters.to_gsa())
            #     # Extract sampled values for parameters, store.
            #     self.param_rng.retrieve_sampled_values(self.parameter_data)

            # self.lca.rebuild_technosphere_matrix(tech_vector)
            # self.lca.rebuild_biosphere_matrix(bio_vector)

            # store matrices for GSA
            self.A_matrices.append(self.lca.technosphere_matrix)
            self.B_matrices.append(self.lca.biosphere_matrix)

            # if not hasattr(self.lca, "demand_array"):
            #     self.lca.build_demand_array()
            # self.lca.lci_calculation()

            # pre-calculating CF vectors enables the use of the SAME CF vector for each FU in a given run
            # cf_vectors = {}
            # for m in self.methods:
            #     cf_vectors[m] = (
            #         self.cf_rngs[m].next() if self.include_cfs else self.cf_rngs[m]
            #     )
            #     # store CFs for GSA (in a list defaultdict)
            #     self.CF_dict[m].append(cf_vectors[m])

            # iterate over FUs
            # for row, func_unit in self.rev_fu_index.items():
            #     self.lca.redo_lci(func_unit)  # lca calculation

            #     # iterate over methods
            #     for col, m in self.rev_method_index.items():
            #         self.lca.switch_method(m)
            #         self.lca.rebuild_characterization_matrix(cf_vectors[m])
            #         self.lca.lcia_calculation()
            #         self.results[iteration, row, col] = self.lca.score

            for row, func_unit in self.func_units_multilca.items():
                # self.lca.redo_lci(func_unit)  # lca calculation

                # iterate over methods
                for col, m in enumerate(self.methods):
                    # self.lca.switch_method(m)
                    # self.lca.rebuild_characterization_matrix(cf_vectors[m])
                    # self.lca.lcia_calculation()
                    self.results[iteration, int(row), col] = self.lca.scores[(m, row)]

        log.info(
            f"Monte Carlo LCA: finished {iterations} iterations for {len(self.func_units)} reference flows and "
            f"{len(self.methods)} methods in {np.round(time() - start, 2)} seconds."
        )

    @property
    def func_units_dict(self) -> dict:
        """Return a dictionary of reference flows (key, demand)."""
        return {key: 1 for func_unit in self.func_units for key in func_unit}

    def get_results_by(self, act_key=None, method=None):
        """Get a slice or all of the results.
        - if a method is provided, results will be given for all reference flows and runs
        - if a reference flow is provided, results will be given for all impact categories and runs
        - if a reference flow and impact category is provided, results will be given for all runs of that combination
        - if nothing is given, all results are returned
        """

        if not self.results.any():
            raise ValueError("You need to perform a Monte Carlo Simulation first.")
            return None

        if act_key:
            act_index = self.activity_index.get(act_key)
            log.info(f"Activity key provided: {act_key} {act_index}")
        if method:
            method_index = self.method_index.get(method)
            log.info(f"Method provided: {method} {method_index}")

        if not act_key and not method:
            return self.results
        elif act_key and not method:
            return np.squeeze(self.results[:, act_index, :])
        elif method and not act_key:
            return np.squeeze(self.results[:, :, method_index])
        elif method and act_key:
            return np.squeeze(self.results[:, act_index, method_index])

    def get_results_dataframe(self, act_key=None, method=None, labelled=True):
        """Return a Pandas DataFrame with results for all runs either for
        - all reference flows and a selected impact categories or
        - all impact categories and a selected reference flow.

        If labelled=True, then the activity keys are converted to a human
        readable format.
        """

        if not self.results.any():
            raise ValueError("You need to perform a Monte Carlo Simulation first.")
            return None

        if act_key and method or not act_key and not method:
            raise ValueError("Must provide activity key or method, but not both.")
        data = self.get_results_by(act_key=act_key, method=method)

        if method:
            labels = self.activity_keys
        elif act_key:
            labels = self.methods

        df = pd.DataFrame(data, columns=labels)

        # optionally convert activity keys to human readable output
        if labelled and method:
            df.columns = self.get_labels(df.columns, max_length=20)

        return df

    @staticmethod
    def get_labels(
        key_list, fields: list = None, separator=" | ", max_length: int = None
    ) -> list:
        fields = fields or ["name", "reference product", "location", "database"]
        # need to do this as the keys come from a pd.Multiindex
        acts = (bd.get_activity(key).as_dict() for key in (k for k in key_list))
        translated_keys = [
            separator.join([act.get(field, "") for field in fields]) for act in acts
        ]
        # if max_length:
        #     translated_keys = [wrap_text(k, max_length=max_length) for k in translated_keys]
        return translated_keys


def perform_MonteCarlo_LCA(project="default", cs_name=None, iterations=10):
    """Performs Monte Carlo LCA based on a calculation setup and returns the
    Monte Carlo LCA object."""
    log.info(f"-- Monte Carlo LCA --\n Project: {project} CS: {cs_name}")
    bd.projects.set_current(project)

    # perform Monte Carlo simulation
    mc = MonteCarloLCA(cs_name)
    mc.calculate(iterations=iterations)
    return mc


if __name__ == "__main__":
    mc = perform_MonteCarlo_LCA(project="ei34", cs_name="kraft paper", iterations=15)

    # test the get_results_by() method
    #    print('\nTesting the get_results_by() method')
    act_key = mc.activity_keys[0]
    method = mc.methods[0]
#    print(mc.get_results_by(act_key=act_key, method=method))
#    print(mc.get_results_by(act_key=act_key, method=None))
#    print(mc.get_results_by(act_key=None, method=method))
#    print(mc.get_results_by(act_key=None, method=None))

# testing the dataframe output
#    print(mc.get_results_dataframe(method=mc.methods[0]))
#    print(mc.get_results_dataframe(act_key=mc.activity_keys[0]))



import pickle

from qtpy import QtCore


class ABMimeData(QtCore.QMimeData):

    def setPickleData(self, mimeType, data):
        self.setData(mimeType, pickle.dumps(data))

    def retrievePickleData(self, mimeType):
        data = self.retrieveData(mimeType, bytes)

        if data is None:
            return

        return pickle.loads(data)



from qtpy import QtWidgets
from qtpy.QtCore import Signal, SignalInstance


class ABRadioButtonCollapser(QtWidgets.QWidget):
    """
    Composite that shows different 'views' depending on what radio button is clicked. After initialization you may
    add different options through the add_option method. These are displayed horizontally and only shown when the
    corresponding radiobutton is clicked.
    """
    buttonClicked: SignalInstance = Signal()
    buttonPressed: SignalInstance = Signal()
    buttonReleased: SignalInstance = Signal()
    buttonToggled: SignalInstance = Signal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._options = {}
        self.button_group = QtWidgets.QButtonGroup()

        self.button_group.buttonClicked.connect(lambda: self.buttonClicked.emit())
        self.button_group.buttonPressed.connect(lambda: self.buttonPressed.emit())
        self.button_group.buttonReleased.connect(lambda: self.buttonReleased.emit())
        self.button_group.buttonToggled.connect(lambda: self.buttonToggled.emit())

        self.setLayout(QtWidgets.QVBoxLayout())

    def __getitem__(self, item):
        """Give option name, returns tuple of the corresponding button and view_widget"""
        return self._options[item]

    def addOption(self, name: str, label: str, w: QtWidgets.QWidget):
        """
        Add a collapsible option to the layout.

        Parameters
        ----------
            name : `str`
                Simple name that can be used to identify the option.
            label : `str`
                Label to be shown next to the radio button. Append a * to disable the button, a # to hide the button,
                or a ~ to select the button.
            view : `QtWidgets.QWidget`
                A QWidget that will be shown once the radio button is checked by the user
        """
        # set the view hidden by default
        w.setHidden(True)

        button = QtWidgets.QRadioButton()

        if '*' in label:
            # if * is in the label the radiobutton will be disabled by default
            label = label.replace('*', '')
            button.setDisabled(True)

        if '#' in label:
            # if # is in the label the button will be hidden by default
            label = label.replace('#', '')
            button.setHidden(True)

        if '~' in label:
            # if ~ is in the label the button wil be set as the default option
            label = label.replace('~', '')
            button.setDefault(True)

        # set up the button
        button.setText(label)
        button.setObjectName(name)
        button.clicked.connect(self.updateCollapse)

        # add the button and view to the correct locations
        self.button_group.addButton(button)
        self._options[name] = (button, w)

        self.layout().addWidget(button)
        self.layout().addWidget(w)

    def hideAll(self, uncheck=True):
        """
        Hides all the views, unchecking all the radio_buttons by default.
        """
        for button, view in self._options.values():
            view.setHidden(True)
        self.button_group.checkedButton().setChecked(not uncheck)

    def updateCollapse(self):
        """
        Slot that check what radio button is checked and only unhides the associated view
        """
        # first hide all
        self.hideAll(uncheck=False)

        # get the name of the checked button
        button_name = self.button_group.checkedButton().objectName()

        # show the associated view
        _, view_widget = self._options[button_name]
        view_widget.setHidden(False)

    def button(self, name: str) -> QtWidgets.QRadioButton:
        """Returns the button associated with the given name"""
        return self._options[name][0]

    def view(self, name: str) -> QtWidgets.QWidget | QtWidgets.QLayout:
        """Returns the view associated with the given name"""
        return self._options[name][1]

    def currentOption(self) -> None | str:
        """Returns the name of the currently checked option. Returns None if nothing is selected"""
        button = self.button_group.checkedButton()
        if not button:
            return None
        return button.objectName()



from functools import partial
from typing import List

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons
from activity_browser.ui.wizards import UncertaintyWizard


class CFUncertaintyModify(ABAction):
    """
    ABAction to launch the UncertaintyWizard for Characterization Factor and handles the output by writing the
    uncertainty data using the ImpactCategoryController to the Characterization Factor in question.
    """

    icon = qicons.edit
    text = "Modify uncertainty"

    @classmethod
    @exception_dialogs
    def run(cls, method_name: tuple, char_factors: List[tuple]):
        wizard = UncertaintyWizard(char_factors[0], application.main_window)
        wizard.complete.connect(partial(cls.wizard_done, method_name))
        wizard.show()

    @staticmethod
    def wizard_done(method_name: tuple, cf: tuple, uncertainty: dict):
        """Update the CF with new uncertainty information, possibly converting
        the second item in the tuple to a dictionary without losing information.
        """
        method = bd.Method(method_name)
        method_dict = {cf[0]: cf[1] for cf in method.load()}

        if isinstance(cf[1], dict):
            cf[1].update(uncertainty)
            method_dict[cf[0]] = cf[1]
        else:
            uncertainty["amount"] = cf[1]
            method_dict[cf[0]] = uncertainty

        method.write(list(method_dict.items()))



from logging import getLogger

import bw2data as bd

from activity_browser.actions.base import ABAction, exception_dialogs

log = getLogger(__name__)


class CSDeleteFunctionalUnit(ABAction):
    text = "Delete Functional Unit"

    @staticmethod
    @exception_dialogs
    def run(cs_name: str, indices: list[int]):
        calculation_setup = bd.calculation_setups[cs_name]

        for index in sorted(set(indices), reverse=True):
            del calculation_setup['inv'][index]

        bd.calculation_setups[cs_name] = calculation_setup
        bd.calculation_setups.serialize()



# -*- coding: utf-8 -*-
import hashlib
from typing import Collection
from logging import getLogger

from bw2io.errors import StrategyError
from bw2io.strategies.generic import (format_nonunique_key_error,
                                      link_iterable_by_fields)
from bw2io.utils import DEFAULT_FIELDS, activity_hash

import bw2data as bd
from bw2data.backends import ActivityDataset, sqlite3_lci_db

from ..bwutils.errors import ExchangeErrorValues
from .commontasks import clean_activity_name

log = getLogger(__name__)

TECHNOSPHERE_TYPES = {"technosphere", "substitution", "production"}
BIOSPHERE_TYPES = {"economic", "emission", "natural resource", "social"}

RELINK_FIELDS = (
    "name",
    "database",
    "categories",
    "unit",
    "reference product",
    "location",
)


def relink_exchanges_dbs(data: Collection, relink: dict) -> Collection:
    """Use this to relink exchanges during an actual import."""
    for act in data:
        for exc in act.get("exchanges", []):
            input_key = exc.get("input", ("", ""))
            if input_key[0] in relink:
                new_key = (relink[input_key[0]], input_key[1])
                try:
                    # try and find the new key
                    _ = bd.get_activity(new_key)
                    exc["input"] = new_key
                except ActivityDataset.DoesNotExist as e:
                    raise ValueError(
                        "Cannot relink exchange '{}', key '{}' not found.".format(
                            exc, new_key
                        )
                    ).with_traceback(e.__traceback__)
    return data


def relink_exchanges_with_db(data: list, old: str, new: str) -> list:
    if old == new:
        return _relink_exchanges(data, new)
    for act in data:
        for exc in (
            exc for exc in act.get("exchanges", []) if exc.get("database") == old
        ):
            exc["database"] = new
    return _relink_exchanges(data, new)


def link_exchanges_without_db(data: list, db: str) -> list:
    for act in data:
        for exc in (exc for exc in act.get("exchanges", []) if "database" not in exc):
            exc["database"] = db
    return _relink_exchanges(data, db)


def _relink_exchanges(data: list, other: str) -> list:
    other = bd.Database(other)
    if len(other) == 0:
        raise StrategyError("Cannot link to empty database")
    act = other.random()
    is_technosphere = act.get("type", "process") == "process"
    kind = TECHNOSPHERE_TYPES if is_technosphere else BIOSPHERE_TYPES
    return link_iterable_by_fields(
        data, other=other, kind=kind, fields=RELINK_FIELDS, relink=True
    )


def relink_exchanges_bw2package(data: dict, relink: dict) -> dict:
    """Use this to relink exchanges during an BW2Package import."""
    for key, value in data.items():
        for exc in value.get("exchanges", []):
            input_key = exc.get("input", ("", ""))
            if input_key[0] in relink:
                new_key = (relink[input_key[0]], input_key[1])
                try:
                    # try and find the new key
                    _ = bd.get_activity(new_key)
                    exc["input"] = new_key
                except ActivityDataset.DoesNotExist as e:
                    raise ValueError(
                        "Cannot relink exchange '{}', key '{}' not found.".format(
                            exc, new_key
                        )
                    ).with_traceback(e.__traceback__)
    return data


def rename_db_bw2package(data: dict, old: str, new: str) -> dict:
    """Replace the given `old` database name with the `new`."""

    def swap(x: tuple) -> tuple:
        return new if x[0] == old else x[0], x[1]

    new_data = {}
    keys = list(data.keys())
    for key in keys:
        value = data.pop(key)
        new_key = swap(key)
        if "database" in value and value["database"] == old:
            value["database"] = new
        for exc in value.get("exchanges", []):
            exc["input"] = swap(exc.get("input", ("", "")))
            exc["output"] = swap(exc.get("output", ("", "")))
        new_data[new_key] = value
    return new_data


def relink_exchanges(exchanges: list, candidates: dict, duplicates: dict) -> tuple:
    altered = 0
    remainder = 0
    unlinked_exchanges = {}
    with sqlite3_lci_db.transaction() as transaction:
        try:
            # Only do relinking on external biosphere/technosphere exchanges.
            for i, exc in exchanges:
                # Use the input activity to generate the hash.
                key = activity_hash(exc.input, DEFAULT_FIELDS)
                if key in duplicates:
                    raise StrategyError(
                        format_nonunique_key_error(
                            exc.input, DEFAULT_FIELDS, duplicates[key]
                        )
                    )
                elif key in candidates:
                    exc["input"] = candidates[key]
                    altered += 1
                else:
                    remainder += 1
                    if len(unlinked_exchanges) <= 5:
                        unlinked_exchanges[exc.input] = key
                exc.save()
                if i % 10000 == 0:
                    # Commit changes every 10k exchanges.
                    transaction.commit()
        except (StrategyError, bd.errors.ValidityError) as e:
            log.error(e)
            transaction.rollback()
    return (remainder, altered, unlinked_exchanges)


def relink_exchanges_existing_db(
    db: bd.Database, old: str, other: bd.Database
) -> tuple:
    """Relink exchanges after the database has been created/written.

    This means possibly doing a lot of sqlite update calls.
    """
    if old == other.name:
        log.info("No point relinking to same database.")
        return
    assert db.backend == "sqlite", "Relinking only allowed for SQLITE backends"
    assert other.backend == "sqlite", "Relinking only allowed for SQLITE backends"

    duplicates, candidates = {}, {}

    for ds in other:
        key = activity_hash(ds, DEFAULT_FIELDS)
        if key in candidates:
            duplicates.setdefault(key, []).append(ds)
        else:
            candidates[key] = ds.key

    exchanges = [
        (i, exc)
        for i, exc in enumerate(
            exc
            for act in db
            for exc in act.exchanges()
            if exc.get("type") in {"biosphere", "technosphere"} and exc.input[0] == old
        )
    ]

    # Process the database after the transaction is complete.
    #  this updates the 'depends' in metadata
    (remainder, altered, unlinked_exchanges) = relink_exchanges(
        exchanges, candidates, duplicates
    )
    db.process()
    log.info(
        "Relinked database '{}', {} exchange inputs changed from '{}' to '{}'.".format(
            db.name, altered, old, other.name
        )
    )
    return (remainder, altered, unlinked_exchanges)


def relink_activity_exchanges(act, old: str, other: bd.Database) -> tuple:
    if old == other.name:
        log.info("No point relinking to same database.")
        return
    db = bd.Database(act.key[0])
    assert db.backend == "sqlite", "Relinking only allowed for SQLITE backends"
    assert other.backend == "sqlite", "Relinking only allowed for SQLITE backends"

    duplicates, candidates = {}, {}

    for ds in other:
        key = activity_hash(ds, DEFAULT_FIELDS)
        if key in candidates:
            duplicates.setdefault(key, []).append(ds)
        else:
            candidates[key] = ds.key
    exchanges = [
        (i, e)
        for i, e in enumerate(
            exc
            for exc in act.exchanges()
            if exc.get("type") in {"technosphere", "biosphere"} and exc.input[0] == old
        )
    ]

    (remainder, altered, unlinked_exchanges) = relink_exchanges(
        exchanges, candidates, duplicates
    )
    db.process()
    log.info(
        "Relinked database '{}', {} exchange inputs changed from '{}' to '{}'.".format(
            db.name, altered, old, other.name
        )
    )
    return (remainder, altered, unlinked_exchanges)


def alter_database_name(data: list, old: str, new: str) -> list:
    """For ABExcelImporter, go through data and replace all instances
    of the `old` database name with `new`.
    """
    if old == new:
        return data  # Avoid doing any work if the two are equal.
    for ds in data:
        # Alter db on activities.
        ds["database"] = new
        for exc in ds.get("exchanges", []):
            # Note: this will only alter database if the field exists in the exchange.
            if exc.get("database") == old:
                exc["database"] = new
        for p, d in ds.get("parameters", {}).items():
            # Any parameters found here are activity parameters and we can
            # overwrite the database without issue.
            d["database"] = new
    return data


def hash_parameter_group(data: list) -> list:
    """For ABExcelImporter, go through `data` and change all the activity parameter
    `group` fields to use a md5 hash instead of the given group name.
    """
    for ds in (ds for ds in data if "parameters" in ds):
        key = (ds.get("database"), ds.get("code"))
        simple_hash = hashlib.md5(":".join(key).encode()).hexdigest()
        clean = clean_activity_name(ds.get("name"))
        for p, d in ds.get("parameters", {}).items():
            d["group"] = "{}_{}".format(clean, simple_hash)
    return data


def csv_rewrite_product_key(data):
    """Convert exchange 'product' key to a 'reference product' one."""
    for ds in data:
        for exc in (e for e in ds.get("exchanges", []) if "product" in e):
            exc["reference product"] = exc.pop("product")
    return data


def excel_error_detection(data: Collection):
    for ds in data:
        for exc in ds.get("exchanges", []):
            if "amount" not in exc:
                raise ExchangeErrorValues(
                    f"An Exception has occurred for activity {ds['name']}"
                    + "with errors in the exchange flows"
                )
    return data



# -*- coding: utf-8 -*-
import os
from setuptools import setup

if "VERSION" in os.environ:
    version = os.environ["VERSION"]
elif "PKG_VERSION" in os.environ:
    version = os.environ["PKG_VERSION"]
else:
    version = os.environ.get("GIT_DESCRIBE_TAG", "0.0.0")
    if "-" in version:
        versions = version.split("-")
        version = "{}.post{}".format(versions[0], versions[1])

setup(
    version=version,
    packages=["activity_browser", "activity_browser_beta"],
    license=open("LICENSE.txt").read(),
    include_package_data=True,
)



# -*- coding: utf-8 -*-
import ast
import sys
from typing import List, Tuple

import numpy as np
import pandas as pd
import bw2data as bd
import bw_functional as bf
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QApplication, QPushButton

from ..errors import ScenarioDatabaseNotFoundError
from ..metadata import AB_metadata
from ..utils import Index
from .activities import data_from_index
from .file_dialogs import ABPopup
from .utils import SUPERSTRUCTURE


def superstructure_from_arrays(
    samples: np.ndarray, indices: np.ndarray, names: List[str] = None
) -> pd.DataFrame:
    """Process indices into the superstructure itself, the samples represent
    the scenarios.
    """
    assert samples.ndim == 2, "Samples array should be 2-dimensional"
    assert indices.ndim == 1, "Indices array should be 1-dimensional"
    assert samples.shape[0] == indices.shape[0], "Length mismatch between arrays"
    if names is not None:
        assert (
            len(names) == samples.shape[1]
        ), "Number of names should match number of samples columns"
        names = pd.Index(names)
    else:
        names = pd.Index(["scenario{}".format(i + 1) for i in range(samples.shape[1])])

    # Construct superstructure from indices
    superstructure = pd.DataFrame(map(data_from_index, indices), columns=SUPERSTRUCTURE)
    # Construct scenarios from samples
    scenarios = pd.DataFrame(samples, columns=names)

    df = pd.concat([superstructure, scenarios], axis=1)
    return df


def superstructure_from_scenario_exchanges(scenarios: dict[str, dict[int, float]]):
    from activity_browser.bwutils import exchanges_to_sdf
    from bw2data import Edge

    scenarios = transpose_scenarios_to_exchange_ids(scenarios)
    df = pd.DataFrame(columns=SUPERSTRUCTURE)

    for exchange_id, amounts in scenarios.items():
        if exchange_is_from_functional_backend(exchange_id):
            df = pd.concat([df, mf_exchange_to_sdf(exchange_id, amounts)], ignore_index=True)
        else:
            df = pd.concat([df, regular_exchange_to_sdf(exchange_id, amounts)], ignore_index=True)

    return df


def regular_exchange_to_sdf(exchange_id: int, scenarios: dict[str, float]):
    from activity_browser.bwutils import exchanges_to_sdf

    exc = bd.Edge(bd.Edge.ORMDataset.get_by_id(exchange_id)).as_dict()
    df = exchanges_to_sdf([exc])
    df.drop(columns=["amount"], inplace=True)
    for scenario, amount in scenarios.items():
        df[scenario] = amount

    return df


def mf_exchange_to_sdf(exchange_id: int, scenarios: dict[str, float]):
    from activity_browser.bwutils import exchanges_to_sdf

    exc = bf.MFExchange(bf.MFExchange.ORMDataset.get_by_id(exchange_id))

    df = exchanges_to_sdf(exc.virtual_edges)
    df.drop(columns=["amount"], inplace=True)
    for scenario, amount in scenarios.items():
        exc["amount"] = amount
        df[scenario] = [edge["amount"] for edge in exc.virtual_edges]

    return df


def exchange_is_from_functional_backend(exchange_id: int):
    db = bd.Edge.ORMDataset.get_by_id(exchange_id).output_database
    return bd.databases[db].get("backend") == "functional_sqlite"


def transpose_scenarios_to_exchange_ids(scenarios: dict[str, dict[int, float]]) -> dict[int, dict[str, float]]:
    """
    Transpose a dictionary of scenarios to organize data by exchange IDs.

    Parameters
    ----------
    scenarios : dict[str, dict[int, float]]
        A dictionary where keys are scenario names (str) and values are dictionaries
        mapping exchange IDs (int) to amounts (float).

    Returns
    -------
    dict[int, dict[str, float]]
        A dictionary where keys are exchange IDs (int) and values are dictionaries
        mapping scenario names (str) to amounts (float).
    """
    scenarios_by_exchange = {}

    for scenario, exchanges in scenarios.items():
        for exchange_id, amount in exchanges.items():
            if exchange_id not in scenarios_by_exchange:
                scenarios_by_exchange[exchange_id] = {}
            scenarios_by_exchange[exchange_id][scenario] = amount

    return scenarios_by_exchange


def arrays_from_indexed_superstructure(
    df: pd.DataFrame,
) -> Tuple[np.ndarray, np.ndarray]:
    result = np.zeros(df.shape[0], dtype=object)
    for i, data in enumerate(df.index.to_flat_index()):
        result[i] = Index.build_from_dict(
            {"input": data[0], "output": data[1], "flow type": data[2]}
        )
    return result, df.to_numpy(dtype=float)


def filter_databases_indexed_superstructure(
    df: pd.DataFrame, include: set
) -> pd.DataFrame:
    """Filters the given superstructure so that only indexes where the output
    database is in the `include` set are valid.
    """
    return df.loc[[x[1][0] in include for x in df.index.to_flat_index()], :]


def scenario_columns(df: pd.DataFrame) -> pd.Index:
    # 'sort=False' ensures that order of scenario names is not altered.
    return df.columns.difference(SUPERSTRUCTURE, sort=False)


def scenario_names_from_df(df: pd.DataFrame) -> List[str]:
    """Returns the list of scenario names from a given superstructure.

    Strip out any possible carriage returns (excel junk)
    """
    cols = scenario_columns(df)
    return [str(x).replace("\n", " ").replace("\r", "") for x in cols]


def scenario_replace_databases(df_: pd.DataFrame, replacements: dict) -> pd.DataFrame:
    """For a provided dataframe the function will check for the presence of a unidentified database for all rows.
    If an unidentified database is found as a key in the replacements argument the corresponding value provided is used
    to provide an alternative database. The corresponding key for the activity from the unidentified database is
    collected from the provided alternative.

    If an activity cannot be identified within the provided database a warning message is provided. The process can
    either be terminated, or can proceed without replacement of those activities not identified (the unidentified
    database names in these instances will be retained)

    Raises
    ------
    ScenarioDatabaseNotFoundError

    Parameters
    ----------

    df_ : the dataframe that is produced from the supplied scenario files provided to the AB

    replacements : a dictionary of key-value pairs where the key corresponds to the database in the supplied dataframe
            and the value corresponds to the respective database in the local brightway environment

    bw_dbs : a list of Brightway databases held locally

    Returns
    -------
    """

    def exchange_replace_database(
        ds: pd.Series, replacements: dict, critical: list, idx: pd.Index
    ) -> tuple:
        """
        For a row in the scenario dataframe check the databases involved for whether replacement is required.
        If so use the key-value pair within the replacements dictionary to replace the dictionary names
        and obtain the correct activity key

        Raises
        ------
        No exception however it creates a store of five values that indicate non-linkable flows with the
        new database

        Parameters
        ----------
        ds: dataseries from a pandas dataframe containing the data from the scenario difference file
        replacements: a key -- value pair containing the old -- new database names
        critical: an initially empty list that is filled with dataseries that fail in the relinking process
        idx: the index for the dataseries object, in the "parent" dataframe
        """
        ds_ = ds.copy()
        for i, field in enumerate([FROM_FIELDS, TO_FIELDS]):
            db_name = ds_[["from database", "to database"][i]]
            # check for the relevance of the particular field
            if db_name not in replacements.keys():
                continue
            try:
                # try to find the matching records (after loaded into the metadata)
                if isinstance(ds_[field[1]], float):
                    # try to find a technosphere record
                    key = metadata[
                        (metadata[DB_FIELDS[0]] == ds_[field[0]])
                        & (metadata[DB_FIELDS[2]] == ds_[field[2]])
                        & (metadata[DB_FIELDS[3]] == ds_[field[3]])
                    ].copy()
                else:
                    # try to find a biosphere record
                    if isinstance(ds_[field[1]], str):
                        categories = ast.literal_eval(ds_[field[1]])
                    else:
                        categories = ds_[field[1]]
                    key = metadata[
                        (metadata[DB_FIELDS[0]] == ds_[field[0]])
                        & (metadata[DB_FIELDS[1]] == categories)
                    ].copy()
                # replace the records that can be found
                for j, col in enumerate(
                    [["from key", "from database"], ["to key", "to database"]][i]
                ):
                    ds_.loc[col] = (
                        (key["database"].iloc[0], key["code"].iloc[0])
                        if j == 0
                        else key["database"].iloc[0]
                    )
            except Exception as e:
                # if the record cannot be found add an exception (to a maximum of five)
                if len(critical["from database"]) <= 5:
                    critical["index"].append(idx)
                    critical["from database"].append(ds_["from database"])
                    critical["from activity name"].append(ds_["from activity name"])
                    critical["to database"].append(ds_["to database"])
                    critical["to activity name"].append(ds_["to activity name"])
        return ds_

    # Create a new database from those records in the scenario files that include exchanges where a replacement database
    # is required
    df = df_.loc[
        (df_["from database"].isin(replacements.keys()))
        | (df_["to database"].isin(replacements.keys()))
    ].copy(True)

    # A LIST OF FIELDS FOR ITERATION
    FROM_FIELDS = pd.Index(
        [
            "from activity name",
            "from categories",
            "from reference product",
            "from location",
        ]
    )
    TO_FIELDS = pd.Index(
        ["to activity name", "to categories", "to reference product", "to location"]
    )
    DB_FIELDS = ["name", "categories", "reference product", "location"]

    # setting up the variables in case some exchanges cannot be relinked
    critical = {
        "index": [],
        "from database": [],
        "from activity name": [],
        "to database": [],
        "to activity name": [],
    }  # To be used in the exchange_replace_database internal method scope
    changes = ["from database", "from key", "to database", "to key"]

    # Load all required databases into the metadata
    AB_metadata.add_metadata(replacements.values())
    metadata = AB_metadata.dataframe

    for idx in df.index:
        df.loc[idx, changes] = exchange_replace_database(
            df.loc[idx, :], replacements, critical, idx
        )[changes]
        sys.stdout.write(
            "\r{}".format(idx / df.shape[0])
        )  # TODO check adaptation for the logger
        sys.stdout.flush()

    if critical["from database"]:
        # prepare a warning message in case unlinkable activities were found in the scenario dataframe
        QApplication.restoreOverrideCursor()
        if len(critical["from database"]) > 1:
            msg = (
                f'Multiple activities could not be "relinked" to the local database.<br> The first five are provided. '
                f"If you want to save the dataframe you can either save those scenario exchanges where relinking failed "
                f"(check the excerpt box), or save the entire dataframe with a new column indicating failed relinking."
                f"<br> To abort the process press 'Cancel'"
            )
            critical_message = ABPopup.abCritical(
                "Activities not found",
                msg,
                QPushButton("Save"),
                QPushButton("Cancel"),
                default=2,
            )
            critical_message.save_options()
            critical_message.dataframe(df.loc[critical["index"], :], SUPERSTRUCTURE)
            critical_message.dataframe_to_file(df_, critical["index"])
            response = critical_message.exec_()
        else:
            msg = (
                f'An activity could not be "relinked" to the local database.<br> Some additional information is '
                f"provided. If you want to save the dataframe you can either save those scenario exchanges where "
                f"relinking failed (check the excerpt box), or save the entire dataframe with a new column indicating"
                f" failed relinking.<br>To abort the process press 'Cancel'"
            )
            critical_message = ABPopup.abCritical(
                "Activity not found",
                msg,
                QPushButton("Save"),
                QPushButton("Cancel"),
                default=2,
            )
            critical_message.save_options()
            critical_message.dataframe(df.loc[critical["index"], :], SUPERSTRUCTURE)
            critical_message.dataframe_to_file(df_, critical["index"])
            response = critical_message.exec_()
        QApplication.setOverrideCursor(Qt.WaitCursor)
        raise ScenarioDatabaseNotFoundError(
            "Incompatible DatabasesPane in the scenario file, unable to complete further checks on the file"
        )
    else:
        df_.loc[df.index] = df
    return df_



from logging import getLogger

from qtpy import QtGui, QtWidgets, QtCore, PYSIDE6
from qtpy.QtCore import Qt
from qtpy.QtGui import QFontDatabase

from activity_browser.static import fonts

log = getLogger(__name__)


class ABApplication(QtWidgets.QApplication):
    _main_window = None
    _controllers = None

    windows = []

    def __init__(self, *args, **kwargs):
        QtCore.QCoreApplication.setAttribute(Qt.AA_ShareOpenGLContexts, True)
        QtCore.QCoreApplication.setAttribute(Qt.AA_UseSoftwareOpenGL)

        super().__init__(*args, **kwargs)

        QtGui.QGuiApplication.setAttribute(Qt.AA_DontShowIconsInMenus, True)

        self.add_fonts()

        if PYSIDE6:
            self.pyside6_setup()

    def add_fonts(self):
        QFontDatabase.addApplicationFont(fonts.__path__[0] + "/mono.ttf")
        QFontDatabase.addApplicationFont(fonts.__path__[0] + "/ptsans.ttf")
        QFontDatabase.addApplicationFont(fonts.__path__[0] + "/notosans.ttf")

    def pyside6_setup(self):
        from qtpy.QtWebEngineQuick import QtWebEngineQuick
        QtWebEngineQuick.initialize()

        style = QtWidgets.QStyleFactory().create("fusion")
        self.setStyle(style)

        self.styleHints().colorSchemeChanged.connect(self.check_palette)
        self.check_palette(self.styleHints().colorScheme())

    def check_palette(self, color_scheme):
        import matplotlib.pyplot as plt

        if color_scheme == Qt.ColorScheme.Dark:
            palette = self.style().standardPalette()
            palette.setColor(QtGui.QPalette.AlternateBase, QtGui.QColor(53, 53, 53))

            plt.style.use("dark_background")

            # os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = "--force-dark-mode"
        else:
            palette = self.style().standardPalette()

            plt.style.use("default")

            # os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = ""
        self.setPalette(palette)

    @property
    def main_window(self) -> QtWidgets.QMainWindow:
        """Returns the main_window widget of the Activity Browser"""
        if self._main_window:
            return self._main_window
        raise Exception(
            "main_window not yet initialized, did you try to access it during startup?"
        )

    @main_window.setter
    def main_window(self, widget: QtWidgets.QMainWindow):
        self._main_window = widget

    def show(self):
        self.main_window.showMaximized()

    def close(self):
        for child in self.children():
            if hasattr(child, "close"):
                child.close()

    def deleteLater(self):
        self.main_window.deleteLater()


application = ABApplication()



#
# if QSysInfo.productType() == "osx":
#     # https://bugreports.qt.io/browse/QTBUG-87014
#     # https://bugreports.qt.io/browse/QTBUG-85546
#     # https://github.com/mapeditor/tiled/issues/2845
#     # https://doc.qt.io/qt-5/qoperatingsystemversion.html#MacOSBigSur-var
#     supported = {"10.10", "10.11", "10.12", "10.13", "10.14", "10.15", "13.6"}
#     if QSysInfo.productVersion() not in supported:
#         os.environ["QT_MAC_WANTS_LAYER"] = "1"
#         os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = "--disable-gpu"
#         log.info("Info: GPU hardware acceleration disabled")
#
# # on macos buttons silently crashes the renderer without any logs
# # confirmed that buttons works on the latest version of qt using pyside6
# if QSysInfo.productType() in ["arch", "nixos", "osx"]:
#     os.environ["QTWEBENGINE_CHROMIUM_FLAGS"] = "{} --no-sandbox".format(
#         os.getenv("QTWEBENGINE_CHROMIUM_FLAGS")
#     )
#     log.info("Info: QtWebEngine sandbox disabled")



# -*- coding: utf-8 -*-
from qtpy import QtCore, QtWidgets
from stats_arrays import uncertainty_choices as uc

from activity_browser import actions

from activity_browser.signals import signals


class UncertaintyDelegate(QtWidgets.QStyledItemDelegate):
    """A combobox containing the sorted list of possible uncertainties
    `setModelData` stores the integer id of the selected uncertainty
    distribution.
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        uc.check_id_uniqueness()
        self.choices = {u.description: u.id for u in uc.choices}

    def displayText(self, value, locale):
        """Take the given integer id and return the description.

        Will return the 'Unknown' uncertainty description if the given id
        either cannot be found or the value is 'nan' (when id is not set)
        """
        try:
            return uc[int(value)].description
        except (IndexError, ValueError):
            return uc[0].description

    def createEditor(self, parent, option, index):
        """Simply use the wizard for updating uncertainties. Send a signal."""
        if hasattr(self.parent(), "modify_uncertainty_action"):
            self.parent().modify_uncertainty_action.trigger()
        elif hasattr(index.internalPointer(), "exchange"):
            item = index.internalPointer()
            actions.ExchangeUncertaintyModify.run([item.exchange])

    def setEditorData(self, editor: QtWidgets.QComboBox, index: QtCore.QModelIndex):
        """Simply use the wizard for updating uncertainties."""
        pass

    def setModelData(
        self,
        editor: QtWidgets.QComboBox,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ):
        """Read the current text and look up the actual ID of that uncertainty type."""
        uc_id = self.choices.get(editor.currentText(), 0)
        model.setData(index, uc_id, QtCore.Qt.EditRole)



from qtpy import QtCore, QtWidgets
from qtpy.QtGui import QFontMetrics, QFont
from qtpy.QtCore import Qt


class NewFormulaDelegate(QtWidgets.QStyledItemDelegate):
    """For managing and validating entered float values."""

    def sizeHint(self, option, index):
        size = super().sizeHint(option, index)

        h = QFontMetrics(QFont("JetBrains Mono", 10)).height() + 4
        size.setHeight(h)  # Set your custom row height here
        return size

    def displayText(self, value, locale):
        return f"<b>{value}</b>"

    def paint(self, painter, option: QtWidgets.QStyleOptionViewItem, index):
        if index.data() is None:
            return super().paint(painter, option, index)

        painter.save()

        if option.state & QtWidgets.QStyle.State_Selected:
            painter.fillRect(option.rect, option.palette.color(option.palette.ColorRole.Highlight))
            painter.setPen(option.palette.color(option.palette.ColorRole.HighlightedText))
        else:
            painter.setPen(Qt.NoPen)

        if hasattr(index.internalPointer(), 'scoped_parameters'):
            scope = index.internalPointer().scoped_parameters
        else:
            scope = {}

        from activity_browser.ui.widgets import ABFormulaEdit
        viewport = self.parent().findChild(QtWidgets.QWidget, "qt_scrollarea_viewport")
        formula = ABFormulaEdit(viewport, scope, index.data())

        painter.setClipRect(option.rect)
        painter.translate(option.rect.topLeft())

        formula.setGeometry(option.rect)
        formula.paint_text(painter)

        painter.restore()

    def createEditor(self, parent, option, index):
        from activity_browser.ui.widgets import ABFormulaEdit
        if hasattr(index.internalPointer(), 'scoped_parameters'):
            scope = index.internalPointer().scoped_parameters
        else:
            scope = {}
        editor = ABFormulaEdit(parent, scope)
        return editor

    def setEditorData(self, editor, index: QtCore.QModelIndex):
        """Populate the editor with data if editing an existing field."""
        value = index.data(QtCore.Qt.DisplayRole)
        # Avoid setting 'None' type value as a string
        value = str(value) if value else ""
        editor.text = value
        editor.cursor_pos = len(value)  # move cursor to the end of the field

    def updateEditorGeometry(self, editor, option, index):
        super().updateEditorGeometry(editor, option, index)

    def setModelData(
        self,
        editor: QtWidgets.QLineEdit,
        model: QtCore.QAbstractItemModel,
        index: QtCore.QModelIndex,
    ):
        model.setData(index, editor.text, QtCore.Qt.EditRole)



import datetime

from qtpy import QtWidgets, QtGui
from qtpy.QtCore import Qt

import bw2data as bd
import pandas as pd

from activity_browser import signals, actions, bwutils
from activity_browser.settings import project_settings
from activity_browser.ui import widgets, icons, delegates


class DatabasesPane(widgets.ABAbstractPane):
    """
    A widget that displays the databases and their details.

    Attributes:
        view (DatabasesView): The view displaying the databases.
        model (DatabasesModel): The model containing the data for the databases.
    """
    title = "Databases"
    unique = True

    def __init__(self, parent):
        """
        Initializes the DatabasesPane widget.

        Args:
            parent (QtWidgets.QWidget): The parent widget.
        """
        super().__init__(parent)
        self.view = DatabasesView()
        self.model = DatabasesModel()
        self.view.setModel(self.model)

        self.view.setAlternatingRowColors(True)
        # self.view.setSelectionMode(QtWidgets.QTableView.SingleSelection)
        self.view.setIndentation(0)

        self.build_layout()
        self.connect_signals()

    def connect_signals(self):
        """
        Connects the signals to the appropriate slots.
        """
        signals.meta.databases_changed.connect(self.sync)
        signals.project.changed.connect(self.sync)
        signals.database.deleted.connect(self.sync)
        signals.database_read_only_changed.connect(self.sync)

    def build_layout(self):
        """
        Builds the layout of the widget.
        """
        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(self.view)
        layout.setContentsMargins(5, 0, 5, 5)
        self.setLayout(layout)

    def sync(self):
        """
        Synchronizes the model with the current state of the databases.
        """
        self.model.setDataFrame(self.build_df())
        self.view.resizeColumnToContents(0)
        self.view.header().setSectionResizeMode(0, QtWidgets.QHeaderView.Fixed)

    def build_df(self) -> pd.DataFrame:
        """
        Builds a DataFrame from the databases.

        Returns:
            pd.DataFrame: The DataFrame containing the databases data.
        """
        data = []
        for name in bd.databases:
            # get the modified time, in case it doesn't exist, just write 'now' in the correct format
            dt = bd.databases[name].get("modified", datetime.datetime.now().isoformat())
            dt = datetime.datetime.strptime(dt, "%Y-%m-%dT%H:%M:%S.%f")

            # final column includes interactive checkbox which shows read-only state of db
            data.append(
                {
                    "name": name,
                    "depends": ", ".join(bd.databases[name].get("depends", [])),
                    "modified": dt,
                    "records": bwutils.commontasks.count_database_records(name),
                    "read_only": bd.databases[name].get("read_only", True),
                    "default_allocation": bd.databases[name].get("default_allocation", "unspecified"),
                    "backend": bd.databases[name].get("backend")
                }
            )

        cols = ["read_only", "name", "records", "depends", "default_allocation", "modified", "backend"]

        return pd.DataFrame(data, columns=cols)


class DatabasesView(widgets.ABTreeView):
    """
    A view that displays the databases in a tree structure.

    Attributes:
        defaultColumnDelegates (dict): The default column delegates for the view.
    """
    defaultColumnDelegates = {
        "modified": delegates.DateTimeDelegate,
    }

    class ContextMenu(widgets.ABMenu):
        menuSetup = [
            lambda m, p: m.add(actions.DatabaseNew),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(actions.DatabaseDelete, p.selected_databases[0] if p.selected_databases else None,
                               enable=len(p.selected_databases) == 1),
            lambda m, p: m.add(actions.DatabaseDuplicate, p.selected_databases[0] if p.selected_databases else None,
                               enable=len(p.selected_databases) == 1),
            lambda m, p: m.add(actions.DatabaseProcess, p.selected_databases[0] if p.selected_databases else None,
                               enable=len(p.selected_databases) == 1),
            lambda m: m.addSeparator(),
            lambda m, p: m.add(actions.DatabaseSetReadonly, p.selected_databases[0] if p.selected_databases else None,
                               not m.selected_readonly,
                               enable=len(p.selected_databases) == 1,
                               text="Unlock database" if m.selected_readonly else "Lock database",
                               ),
        ]

        @property
        def selected_readonly(self):
            """
            Returns the read-only state of the selected database.

            Returns:
                bool: The read-only state of the selected database.
            """
            if not self.parent().selected_databases:
                return None
            return self.parent().selectedIndexes()[0].internalPointer()["read_only"]

    class HeaderMenu(QtWidgets.QMenu):
        """
        A header menu for the DatabasesView. Currently not used.
        """

        def __init__(self, *args, **kwargs):
            super().__init__()

    def mouseDoubleClickEvent(self, event: QtGui.QMouseEvent):
        """
        Handles the mouse double click event to toggle the read-only state or select the database.

        Args:
            event (QtGui.QMouseEvent): The mouse double click event.
        """
        index = self.indexAt(event.pos())

        if not index.isValid():
            return super().mouseDoubleClickEvent(event)

        db_name = index.internalPointer()["name"]

        if index.column() == 0:
            read_only = index.internalPointer()["read_only"]
            actions.DatabaseSetReadonly.run(db_name, not read_only)
            return

        actions.DatabaseOpen.run([db_name])

    @property
    def selected_databases(self) -> list:
        """
        Returns the database name of the user-selected index.

        Returns:
            str: The name of the selected database.
        """
        if not self.selectedIndexes():
            return []
        return list(set([i.internalPointer()["name"] for i in self.selectedIndexes()]))


class DatabasesItem(widgets.ABDataItem):
    """
    An item representing a database in the tree view.
    """

    def decorationData(self, col: int, key: str):
        """
        Provides decoration data for the item.

        Args:
            col (int): The column index.
            key (str): The key for which to provide decoration data.

        Returns:
            The decoration data for the item.
        """
        if key == "read_only":
            return icons.qicons.locked if self["read_only"] else icons.qicons.empty
        return super().decorationData(col, key)

    def displayData(self, col: int, key: str):
        """
        Provides display data for the item.

        Args:
            col (int): The column index.
            key (str): The key for which to provide display data.

        Returns:
            The display data for the item.
        """
        if key == "read_only":
            return None
        return super().displayData(col, key)

    def fontData(self, col: int, key: str):
        """
        Provides font data for the item.

        Args:
            col (int): The column index.
            key (str): The key for which to provide font data.

        Returns:
            QtGui.QFont: The font data for the item.
        """
        font = super().fontData(col, key)
        if key == "name":
            font.setWeight(QtGui.QFont.Weight.DemiBold)
        return font


class DatabasesModel(widgets.ABItemModel):
    """
    A model representing the data for the databases.

    Attributes:
        dataItemClass (type): The class of the data items.
    """
    dataItemClass = DatabasesItem

    def headerData(self, section, orientation=Qt.Orientation.Horizontal, role=Qt.ItemDataRole.DisplayRole):
        """
        Provides header data for the model.

        Args:
            section (int): The section index.
            orientation (Qt.Orientation): The orientation of the header.
            role (Qt.ItemDataRole): The role for which to provide header data.

        Returns:
            The header data for the model.
        """
        if section == 0 and role == Qt.ItemDataRole.DisplayRole:
            return ""
        if section == 0 and role == Qt.ItemDataRole.DecorationRole:
            return icons.qicons.unlocked
        return super().headerData(section, orientation, role)



from zipfile import ZipFile

from bw2io.importers.ecospold2_biosphere import *
import pyprind
import logging
import os

from activity_browser.info import __ei_versions__
from activity_browser.utils import sort_semantic_versions


class ABEcospold2BiosphereImporter(Ecospold2BiosphereImporter):

    def extract(self, version: str | None, filepath: Path | None):
        """
        Extract elementary flows from the xml file.

        Parameters
        ----------
        version
            Version of the database if using default data.
        filepath
            File path of user-specified data file

        Returns
        -------
        list
            Extracted data from the xml file.
        """

        def extract_flow_data(o):
            ds = {
                "categories": (
                    o.compartment.compartment.text,
                    o.compartment.subcompartment.text,
                ),
                "code": o.get("id"),
                "CAS number": o.get("casNumber"),
                "synonyms": [
                    elem.text.strip()
                    for elem in o.iterchildren()
                    if elem.tag == "{http://www.EcoInvent.org/EcoSpold02}synonym"
                    and elem.text  # 3.7.1 has blank elements
                    and elem.text.strip()
                ],
                "name": o.name.text,
                "database": self.db_name,
                "exchanges": [],
                "unit": o.unitName.text,
            }
            ds["type"] = EMISSIONS_CATEGORIES.get(
                ds["categories"][0], ds["categories"][0]
            )
            return ds

        if version != '3.9' and not filepath:
            import activity_browser.bwutils as mod
            lci_dirpath = os.path.join(os.path.dirname(mod.__file__), "ecoinvent_biosphere_versions", "legacy_biosphere")

            # find the most recent legacy biosphere that is equal to or older than chosen version
            for ei_version in sort_semantic_versions(__ei_versions__):
                use_version = ei_version
                zip_fp = os.path.join(
                    lci_dirpath, f"ecoinvent elementary flows {use_version}.xml.zip"
                )
                if sort_semantic_versions([version, ei_version])[
                    0
                ] == version and os.path.isfile(zip_fp):
                    # this version is equal/lower and available
                    break

            # extract the xml from the zip
            with ZipFile(zip_fp) as zipped_file:
                with zipped_file.open(
                        f"ecoinvent elementary flows {use_version}.xml"
                ) as file:
                    root = objectify.parse(file).getroot()
        else:
            if not filepath:
                import bw2io.importers.ecospold2_biosphere as mod
                filepath = (
                    Path(mod.__file__).parent.parent.resolve()
                    / "data"
                    / "lci"
                    / f"ecoinvent elementary flows {version}.xml"
                )

            root = objectify.parse(open(filepath, encoding="utf-8")).getroot()

        flow_data = []

        # AB implementation: added prog_bar here
        for ds in pyprind.prog_bar(list(root.iterchildren()), title="Extracting biosphere data"):
            flow_data.append(extract_flow_data(ds))
        return flow_data

    def apply_strategies(self, strategies=None, verbose=True):
        """Apply a list of strategies.

        Uses the default list ``self.strategies`` if ``strategies`` is ``None``.

        Args:
            *strategies* (list, optional): List of strategies to apply. Defaults to ``self.strategies``.

        Returns:
            Nothings, but modifies ``self.data``, and adds each strategy to ``self.applied_strategies``.

        """
        func_list = self.strategies if strategies is None else strategies
        for func in pyprind.prog_bar(func_list, title="Applying strategies"):
            self.apply_strategy(func, verbose)

    def write_database(self, *args, **kwargs):
        logging.getLogger(__name__).info("Writing Biosphere database")
        super().write_database(*args, **kwargs)



import os
import json
import tarfile
from logging import getLogger

from qtpy import QtWidgets, QtCore

from activity_browser import application
from activity_browser.mod import bw2data as bd
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.threading import ABThread

log = getLogger(__name__)


class ProjectExport(ABAction):
    """
    ABAction to export the current project. Prompts the user to return a save-file location. And then start a thread to
    package the project and save it there. Saving code copied from bw2data.backup.
    """
    icon = application.style().standardIcon(QtWidgets.QStyle.SP_DriveHDIcon)
    text = "&Export this project..."
    tool_tip = "Export project to file"

    @staticmethod
    @exception_dialogs
    def run(project_name: str = None):
        """Export the current project to a folder chosen by the user."""
        if project_name is None:
            project_name = bd.projects.current

        # get target path from the user
        save_path, save_type = QtWidgets.QFileDialog.getSaveFileName(
            parent=application.main_window,
            caption="Choose where",
            dir=os.path.expanduser(f"~/{project_name}.tar.gz"),
            filter="Tar-file (*.tar.gz)"
        )

        if not save_path: return

        # setup dialog
        progress = QtWidgets.QProgressDialog(
            parent=application.main_window,
            labelText="Exporting project",
            maximum=0
        )
        progress.setCancelButton(None)
        progress.setWindowTitle("Exporting project")
        progress.setWindowFlag(QtCore.Qt.WindowContextHelpButtonHint, False)
        progress.setWindowFlag(QtCore.Qt.WindowCloseButtonHint, False)
        progress.findChild(QtWidgets.QProgressBar).setTextVisible(False)
        progress.resize(400, 100)
        progress.show()

        thread = ExportThread(application)
        setattr(thread, "save_path", save_path)
        setattr(thread, "project_name", project_name)
        thread.finished.connect(lambda: progress.deleteLater())
        thread.start()


class ExportThread(ABThread):
    save_path: str
    project_name: str

    def run_safely(self):
        project_dir = os.path.join(bd.projects._base_data_dir, bd.utils.safe_filename(self.project_name))

        with open(os.path.join(project_dir, ".project-name.json"), "w") as f:
            json.dump({"name": self.project_name}, f)

        log.info("Creating project backup archive - this could take a few minutes...")
        with tarfile.open(self.save_path, "w:gz") as tar:
            tar.add(project_dir, arcname=bd.utils.safe_filename(self.project_name))

        log.info(f"Project `{self.project_name}` exported.")



from typing import Any

from bw2data.parameters import ParameterBase, parameters

from activity_browser.ui.icons import qicons
from activity_browser.bwutils import refresh_parameter, Parameter
from activity_browser.actions.base import ABAction, exception_dialogs

from .parameter_rename import ParameterRename


class ParameterModify(ABAction):
    """
    ABAction to delete an existing parameter.
    """

    icon = qicons.edit
    text = "Modify Parameter"

    @staticmethod
    @exception_dialogs
    def run(parameter: tuple | Parameter | ParameterBase, field: str, value: any):
        parameter = refresh_parameter(parameter)
        param_model = parameter.to_peewee_model()

        if field == "data":
            param_model.data.update(value)
        elif field == "name":
            return ParameterRename.run(parameter, value)
        elif field in dir(param_model):
            setattr(param_model, field, value)
        else:
            param_model.data.update({field: value})

        param_model.save()

        if field in ("amount", "formula"):
            parameters.recalculate()



from typing import Any, List

import pandas as pd

from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.bwutils import commontasks
from activity_browser.ui.icons import qicons


class ExchangeCopySDF(ABAction):
    """
    ABAction to copy the exchange information in SDF format to the clipboard.
    """

    icon = qicons.superstructure
    text = "Exchanges for scenario difference file"

    @staticmethod
    @exception_dialogs
    def run(exchanges: List[Any]):
        data = commontasks.get_exchanges_in_scenario_difference_file_notation(exchanges)
        df = pd.DataFrame(data)
        df.to_clipboard(excel=True, index=False)



from qtpy import QtWidgets, QtGui, QtCore

from bw2data.project import projects

from activity_browser import application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.ui.icons import qicons
from activity_browser.ui.threading import ABThread


class ProjectMigrate25(ABAction):
    """
    ABAction to duplicate a project. Asks the user for a new name. Returns if no name is given, the user cancels, or
    when the name is already in use by another project. Else, instructs the ProjectController to duplicate the current
    project to the new name.
    """

    icon = qicons.copy
    text = "Migrate project"
    tool_tip = "Migrate the project to bw25"

    @staticmethod
    @exception_dialogs
    def run(name: str = None):
        if name is None:
            name = projects.current

        dialog = MigrateDialog(name, application.main_window)
        dialog.exec_()

        if dialog.result() == dialog.DialogCode.Rejected:
            return

        if name != projects.current:
            projects.set_current(name, update=False)

        # setup dialog
        progress = QtWidgets.QProgressDialog(
            parent=application.main_window,
            labelText="Migrating project, this may take a while...",
            maximum=0
        )
        progress.setCancelButton(None)
        progress.setWindowTitle("Migrating project to Brightway25")
        progress.setWindowFlag(QtCore.Qt.WindowContextHelpButtonHint, False)
        progress.setWindowFlag(QtCore.Qt.WindowCloseButtonHint, False)
        progress.findChild(QtWidgets.QProgressBar).setTextVisible(False)
        progress.resize(400, 100)
        progress.show()

        thread = MigrateThread(application)
        thread.finished.connect(lambda: progress.deleteLater())
        thread.start()


class MigrateDialog(QtWidgets.QDialog):
    def __init__(self, project_name: str, parent=None):
        from .project_export import ProjectExport

        super().__init__(parent)
        self.setWindowTitle("Migrate project")
        label = QtWidgets.QLabel(f"Migrate project ({project_name}) from legacy to Brightway25. This cannot be undone.")

        cancel = QtWidgets.QPushButton("Cancel")
        migrate = QtWidgets.QPushButton("Migrate")
        backup = ProjectExport.get_QButton(project_name)
        backup.setText("Back-up project")
        backup.setIcon(QtGui.QIcon())

        cancel.clicked.connect(self.reject)
        migrate.clicked.connect(self.accept)

        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addWidget(backup)
        button_layout.addStretch()
        button_layout.addWidget(cancel)
        button_layout.addWidget(migrate)

        layout = QtWidgets.QVBoxLayout()
        layout.addWidget(label)
        layout.addLayout(button_layout)

        self.setLayout(layout)


class MigrateThread(ABThread):
    def run_safely(self):
        projects.migrate_project_25()
        projects.set_current(projects.current)




# -*- coding: utf-8 -*-

"""
A series of defined Errors and Warnings for the Activity Browser

Both Warnings and Exceptions are customized to enable custom handling (in bulk) of non-critical
errors.


"""
from bw2data.errors import *
from bw2io.errors import *


class ABError(Exception):
    """To be used as a generic Activity-Browser Error that will not lead to the AB crashing out"""

    pass


class ABWarning(Warning):
    """To be used as a generic Activity-Browser Warning"""

    pass


class ImportCanceledError(ABError):
    """Import of data was cancelled by the user."""

    pass


class LinkingFailed(ABError):
    """Unlinked exchanges remain after relinking."""

    pass


class IncompatibleDatabaseNamingError(ABError):
    """Database and keys do not match."""

    pass


class ActivityProductionValueError(ABError):
    """Production value for an activity == 0"""

    pass


class InvalidSDFEntryValue(ABError):
    """NA values found for data type that cannot hold \"NA\"."""

    pass


class ExchangeErrorValues(ABError):
    """In Brightway2 if there is an error in an exchange calculation the 'amount' field is not available for the
    Exchange"""

    pass


class ScenarioExchangeError(ABError):
    """In the AB we require the exchanges from the scenario file to be mappable to the databases. If this is not the
    case we MUST throw an error."""

    pass


class ReferenceFlowValueError(ABWarning):
    """While a user can technically perform a calculation with the reference flows all set to 0, such a calculation
    makes no logical sense and will lead to downstream errors (due to 0 results)."""

    pass


class DuplicatedScenarioExchangeWarning(ABWarning):
    """Will warn the user that a loaded scenario table contains duplicate exchanges. Only the last added exchange value
    will be used."""

    pass


class CriticalCalculationError(ABError):
    """Should be raised if some action during the running of the calculation causes a critical Exception that will fail
    the calculation. This is intended to be used with a Popup warning system that catches the original exception.
    """

    pass


class CriticalScenarioExtensionError(ABError):
    """Should be raised when combinging multiple scenario files by extension leads to zero scenario columns. Due to no
    scenario columns being found in common between the scenario files."""

    pass


class ScenarioDatabaseNotFoundError(ABError):
    """Should be raised when looking up one of the processes in an SDF file and the values used don't match those
    present in the local AB/BW databases."""

    pass


class ScenarioExchangeNotFoundError(ABError):
    """Should be raised when looking up a process key from the metadata in a scenario difference file, if THAT process
    key cannot be located in the local databases."""

    pass


class ScenarioExchangeDataNotFoundError(ABError):
    """Should be raised if no actual quantities for the exchanges can be found in the scenario difference file"""

    pass


class ScenarioExchangeDataNonNumericError(ABError):
    """Should be raised if non-numeric data is provided for the exchanges in a scenario difference file."""

    pass


class UnalignableScenarioColumnsWarning(ABWarning):
    """Should be raised if there is a mismatch between the scenario columns from multiple scenario difference files"""


class WrongFileTypeImportError(ABError):
    """Should be raised when a user tries to import the wrong type of file for the import in question.
    For example a database file with the scenario import dialog, or vice versa."""

    pass



from qtpy import QtWidgets

from activity_browser import settings, application
from activity_browser.actions.base import ABAction, exception_dialogs
from activity_browser.mod import bw2data as bd
from activity_browser.ui.icons import qicons


class ProjectDelete(ABAction):
    """
    ABAction to delete the currently active project. Return if it's the startup project.
    """

    icon = qicons.delete
    text = "Delete"
    tool_tip = "Delete the project"

    @staticmethod
    @exception_dialogs
    def run(project_names: [str] = None):
        if project_names is None:
            # get the current project
            project_names = [bd.projects.current]

        if len(project_names) == 0:
            return

        # if it's the startup project: reject deletion and inform user
        if settings.ab_settings.startup_project in project_names:
            QtWidgets.QMessageBox.information(
                application.main_window,
                "Not possible",
                "Can't delete the startup project. Please select another startup project in the settings first.",
            )
            return

        # open a delete dialog for the user to confirm, return if user rejects
        delete_dialog = ProjectDeletionDialog(project_names, application.main_window)
        if delete_dialog.exec_() != ProjectDeletionDialog.Accepted:
            return

        # try to delete the project, delete directory if user specified so
        if bd.projects.current in project_names:
            bd.projects.set_current(settings.ab_settings.startup_project)

        for project in project_names:
            bd.projects.delete_project(
                project, delete_dialog.deletion_warning_checked()
            )

        # inform the user of successful deletion
        QtWidgets.QMessageBox.information(
            application.main_window, "Project(s) deleted", "Project(s) successfully deleted"
        )


class ProjectDeletionDialog(QtWidgets.QDialog):

    def __init__(self, projects: [str], parent=None):
        super().__init__(parent)

        self.title = "Confirm project deletion"

        if len(projects) == 1:
            self.label = QtWidgets.QLabel(
                f"Final confirmation to remove project: {projects[0]}.\n"
                + "Warning: Non reversible process!"
            )
        else:
            self.label = QtWidgets.QLabel(
                f"Final confirmation to remove {len(projects)} projects.\n"
                + "Warning: Non reversible process!"
            )
        self.check = QtWidgets.QVBoxLayout()
        self.hd_check = QtWidgets.QCheckBox(f"Remove from the hard disk")
        self.hd_check.setChecked(True)

        self.buttons = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
        )
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

        self.setWindowTitle(self.title)
        self.layout = QtWidgets.QVBoxLayout()
        self.layout.addWidget(self.label)
        self.layout.addWidget(self.hd_check)
        self.layout.addWidget(self.buttons)

        self.setLayout(self.layout)

    def deletion_warning_checked(self):
        return self.hd_check.isChecked()
